import { Injectable } from '@angular/core'
import { Observable, forkJoin, BehaviorSubject } from 'rxjs'
import { map } from 'rxjs/operators'

import { Apollo, gql } from 'apollo-angular'

import { StyleSpecification } from 'maplibre-gl'

import { TreeData, PrimaryGroupNode, LayerNode } from './models'
import { MapControlService } from './map-control.service'
import { MapService, MapInitData } from '../openapi'

// XXX: unused
const baseStyleListQuery = gql`
query base_style_list {
    base_style_list {
        name
    }
}
`

const baseStyleQuery = gql`
query base_style(
    $name: String!
) {
    base_style(name: $name) {
        name
        style
}
}
`

const baseMapQuery = gql`
query baseMap {
    baseMap {
        name
        stores {
            name
        }
    }
}
`

const createBaseMapQuery = gql`
mutation createBaseMap(
    $baseMapName: String!
    $stores: [String]!
) {
    createBaseMap(
        baseMapName: $baseMapName,
        stores: $stores,
    ) {
        baseMap {
            name
            stores {
                name
            }
        }
    }
}
`

const geomGroupQuery = gql`
query geomGroup {
    geomGroup {
        name
        title
        description
    }
}
`

const storeQuery = gql`
query stores {
    stores {
        name
        store
        rawSurveyStore
        style
        zIndex
        count
        type
        gisType
        description
        custom
        group
        icon
        symbol
        live
        tagPlugins
        viewableRole
    }
}
`

const mapInitDataQuery = gql`
query mapInitData {
    mapInitData {
        baseMaps {
            name
            stores {
                name
            }
        }
        baseStyles {
            name
        }
        groups {
            name
            title
            description
        }
        stores {
            name
            store
            rawSurveyStore
            style
            zIndex
            count
            type
            gisType
            description
            custom
            group
            icon
            symbol
            live
            tagPlugins
            viewableRole
        }
    }
}
`

const gisTypeSymbolMap = {
    Point: '\ue32b',
    Line: '\ue32c',
    Polygon: '\ue32d',
}

export class BaseStyle {
    constructor(
        public name: string,
        public style?: StyleSpecification,
    ) {}
}

export class Store {
    constructor(
        public name: string,
    ) {}
}

export class BaseMap {
    constructor(
        public name: string,
        public stores: Store[],
    ) {}

    pprint() {
        return this.stores.map(store => store.name).join(', ')
    }
}

// export class MapInitData {
//     constructor(
//         public baseStyles: BaseStyle[] = [],
//         public baseMaps: BaseMap[] = [],
//         public groups: PrimaryGroupNode[] = [],
//         public stores: LayerNode[] = [],
//     ) {}
// }

@Injectable()
export class MapDataService {
    constructor(
        // private apollo: Apollo,
        protected mapControlService: MapControlService,
        public mapService: MapService,
    ) {
        // Get the data on init, as it is (most probably) immutable
        this.getMapInitData().subscribe(
            data => {
                this.mapInitDataLoaded.next(true)
                this.mapInitDataLoaded.complete()
            }
        )
    }

    public mapInitDataLoaded = new BehaviorSubject<boolean>(false)
    mapInitDataLoaded$ = this.mapInitDataLoaded.asObservable()

    mapInitData: MapInitData

    getMapInitData(): Observable<MapInitData> {
        return this.mapService.getInitDataApiMapInitDataGet().pipe(map(
            data => this.mapInitData = data
        ))
        // return this.apollo.query({
            // query: mapInitDataQuery,
        // ).pipe(map(
        //     mid => {
        //         this.mapInitData = mid
        //         const baseMaps = mid.baseMaps.map(
        //             bm => new BaseMap(
        //                 bm.name,
        //                 bm.stores.map(
        //                     (store: object) => new Store(store['name'])
        //                 )
        //             )
        //         )
        //         const baseStyles = mid.baseStyles.map(
        //             baseStyle => new BaseStyle(baseStyle.name)
        //         )
        //         const groups = mid.groups.map(
        //             gg => new PrimaryGroupNode(
        //                 gg.name,
        //                 gg.title,
        //                 gg.description
        //             )
        //         )
        //         const stores = mid.stores.map(
        //             layer => new LayerNode(
        //                 layer['name'],
        //                 layer['group'],
        //                 layer['icon'],
        //                 layer['symbol'] || gisTypeSymbolMap[layer['gisType']],
        //                 layer['store'],
        //                 layer['rawSurveyStore'],
        //                 layer['type'],
        //                 layer['gisType'],
        //                 layer.style,
        //                 layer['zIndex'],
        //                 layer['count'],
        //                 layer['description'],
        //                 layer['live'],
        //                 layer['custom'],
        //                 layer['tagPlugins'],
        //                 layer['viewableRole']
        //             )
        //         )
        //         return new MapInitData(
        //             baseStyles,
        //             baseMaps,
        //             groups,
        //             stores,
        //         )
        //     }
        // ))
    }

    // getBaseMaps(): Observable<BaseMap[]> {
    //     return this.apollo.query({
    //         query: baseMapQuery,
    //     }).pipe(map(
    //         res => {
    //             let bms: object = res['data']['baseMap']
    //             return bms['map'](
    //                 (bm: object) => new BaseMap(
    //                     bm['name'],
    //                     bm['stores'].map(
    //                         (store: object) => new Store(store['name'])
    //                     )
    //                 )
    //             )
    //         }
    //     ))
    // }

    public createBaseMap(baseMapName: string, stores: string[]): Observable<BaseMap> {
        return this.apollo.mutate({
            mutation: createBaseMapQuery,
            variables: {
                baseMapName: baseMapName,
                stores: stores
            }
        }).pipe(map(
            res => {
                let bm: object = res['data']['createBaseMap']['baseMap']
                return new BaseMap(
                    bm['name'],
                    bm['stores'].map((store: object) => new Store(store['name']))
                )
            }
        ))
    }

    // XXX: unused
    // getBaseStyleList(): Observable<BaseStyle[]> {
    //     return this.apollo.query({query: baseStyleListQuery}).pipe(map(
    //         data => data['data']['base_style_list'].map(
    //             (baseStyle: string) => new BaseStyle(baseStyle['name'])
    //         )
    //     ))
    // }

    getBaseStyle(styleName: string): Observable<BaseStyle> {
        return this.apollo.query({
            query: baseStyleQuery,
            variables: {
                name: styleName
            }
        }).pipe(map(
            data => new BaseStyle(data['data']['base_style']['name'],
                                  JSON.parse(data['data']['base_style']['style']))
        ))
    }

    // getStores(): Observable<LayerNode[]> {
    //     return this.apollo.query({
    //         query: storeQuery,
    //     }).pipe(map(
    //         res => res['data']['stores'].map(
    //             (layer: Object) => new LayerNode(
    //                 layer['name'],
    //                 layer['group'],
    //                 layer['icon'],
    //                 layer['symbol'] || gisTypeSymbolMap[layer['gisType']],
    //                 layer['store'],
    //                 layer['rawSurveyStore'],
    //                 layer['type'],
    //                 layer['gisType'],
    //                 layer['style'],
    //                 layer['zIndex'],
    //                 layer['count'],
    //                 layer['description'],
    //                 layer['live'],
    //                 layer['custom'],
    //                 layer['tagPlugins'],
    //                 layer['viewableRole']
    //             )
    //         )
    //     ))
    // }

    // getPrimaryGroups(): Observable<PrimaryGroupNode[]> {
    //     return this.apollo.query({
    //         query: geomGroupQuery,
    //     }).pipe(map(
    //         res => res['data']['geomGroup'] && res['data']['geomGroup'].map(
    //             (gg: Object) => new PrimaryGroupNode(
    //                 gg['name'],
    //                 gg['title'],
    //                 gg['description']
    //             )
    //         )
    //     ))
    // }
}